"""This module contains custom filters for the translation from mkdocs to sphinx."""

from jinja2 import contextfilter


def add_filters(app):
    """Register custom filters for page rendering.

    Args:
        app: The Sphinx application.
    """
    app.builder.templates.environment.filters["url"] = url_filter


@contextfilter
def url_filter(context, path):
    """Returns the path to static content allowing for the 'assets/' path used by mkdocs.

    If the path is not to the 'assets/' directory, the given path is returned unchanged.

    Args:
        context (dict): The Jinja context for the page being rendered. This is
            given by Jinja due to the use of the @contextfilter decorator.
        path (str): A path to static (or other) content.

    Returns:
        str: A path to the content accounting for the mkdocs static content path of 'assets/'.
    """
    if "assets/" in path and "pathto" in context:
        path = path.replace("assets/", "_static/")
        path = context["pathto"](path, 1)

    return path



"""Functions to inject additional context for the translation from mkdocs to sphinx."""

import datetime
from collections import defaultdict

import sphinx
import sphinx.util.logging

from sphinx_bluebrain_theme.utils.search_builder import SearchIndexBuilder
from sphinx_bluebrain_theme.utils.navutils import Page


logger = sphinx.util.logging.getLogger(__name__)


def build_adjacent_page(data):
    """Builds an adjacent page with the minimum required data."""
    if not data:
        return None

    page = Page()
    page.title = data.get("title", "")
    page.url = data.get("link", "")

    return page


def build_page(con, pagename):
    """Builds a mkdocs compatible page object and returns it.

    Args:
        con (dict): A dictionary of the context used to build the page.
        pagename (str): The current page name.

    Returns:
        A dictionary representing the page.
    """
    # add the page context variable
    page = Page()

    # Ensure this is set before .parents otherwise it will try to set recursively.
    page.active = True

    page.previous_page = build_adjacent_page(con["prev"])
    page.next_page = build_adjacent_page(con["next"])
    page.toc = con["localtoc"] or []
    page.url = pagename + ".html"
    page.edit_url = None
    page.content = con["body"] or ""
    page.meta = con["meta"]
    page.title = con["title"]
    page.canonical_url = con["pageurl"]
    page.ancestors = con["parents"]

    pm = page.meta

    if pm:
        if "authors" in pm:
            pm["authors"] = pm["authors"].split(",")
        if "source" in pm:
            pm["source"] = pm["source"].split(",")
        if "redirect" in pm:
            pm["redirect"] = con["pathto"](pm["redirect"])

    page.is_homepage = con["pagename"] == con["master_doc"]

    return page


def build_font(con):
    """Build font from context."""
    fonts = {"text": con["theme_font_text"], "code": con["theme_font_code"]}
    font = con["theme_use_google_fonts"] is True

    if font and any(fonts.values()):
        font = {"text": fonts["text"] or "", "code": fonts["code"] or ""}

    return font


def build_extra(con, lang):
    """Build the extra from context."""
    extra = {
        "repo_icon": con["theme_repo_icon"],
        "social": con["theme_social"],
        "disqus": None,  # disqus not currently supported
        "manifest": con["theme_manifest"],
    }

    # default tokenizer for english splits on hypens,
    # preserve them instead
    if lang == "en":
        extra["search"] = {"tokenizer": r"[\s]+"}

    return extra


def build_config(con):
    """
    This fuctions builds a mkdocs compatible config entry.

    Args:
        con (dict): A dictionary of the context used to build the page.

    Returns:
        A dictionary representing the config for the page.
    """
    # do this first so we can set the search tokenizer in
    # extra.search.tokenizer
    lang = con["language"] or "en"

    # add config context variable
    extra = build_extra(con, lang)
    font = build_font(con)

    config = {
        "extra_css": [],
        "site_name": con["docstitle"],
        "theme": {
            "feature": {"tabs": con["theme_feature_tabs"]},
            "palette": {
                "primary": con["theme_palette_primary"],
                "accent": con["theme_palette_accent"],
            },
            "font": font,
            "language": lang,
            # note we use 'assets/' below since the urls use the url filter
            # which will automatically clean up and convert '_static/'
            "logo": "assets/" + (con["logo"] or con["theme_logo"])
            if con["logo"] or con["theme_logo"]
            else {"icon": con["theme_logo_icon"]},
            "favicon": "assets/" + con["favicon"] if con["favicon"] else "",
        },
        "repo_url": con["theme_repo_url"],
        "repo_name": con["theme_repo_name"],
        "copyright": con["copyright"],
        "title_url": con["pathto"](con["master_doc"]),
        "extra": extra,
        "site_url": con["theme_logo_url"] or con["pathto"](con["master_doc"]),
        "site_author": con["theme_site_author"],
        "site_description": con["theme_site_description"],
        "plugins": {"search"},  # need to include for search bar to show
    }

    return config


# pylint: disable=unused-argument
def inject_context_variables(app, pagename, templatename, context, doctree):
    """This function injects the additional context for the translation from mkdocs to sphinx.

    Args:
        app: The Sphinx application.
        pagename (str): The name of the current page.
        templatename (str): The name of the template used to render the page.
        context (dict): The context passed to the page template for rendering.
        doctree: A doctree when the page is created from reST documents.
    """
    con = defaultdict(lambda: None)
    con.update(context)

    extra_context = {}

    # build the page
    extra_context["page"] = build_page(con, pagename)

    # build the config context entry
    extra_context["config"] = build_config(con)

    # fix the nav variable
    extra_context["nav"] = con["globaltoc"]

    # inject new context
    context.update(extra_context)

    # cleanup some defaults which are useful for BBP
    bbp_context_cleanup(context)

    # handle adding the page to the search index
    search_index = app.builder.globalcontext.get("search_index", SearchIndexBuilder())
    search_index.add_page_to_index(context["page"])
    app.builder.globalcontext["search_index"] = search_index

    # remove the "title" context variable as it is now in page.title
    # and conflicts with nav_item.title template code
    if "title" in context:
        del context["title"]


def bbp_context_cleanup(context):
    """Cleans up some theme defaults which are required for BBP projects.

    theme.conf variables are all strings, but templates expect
    other types to be passed (lists and dicts).

    Args:
        context (dict): The context passed to the page template for rendering.
    """
    # do some convenience updates for BBP defaults
    if isinstance(context["theme_address"], str):
        context["theme_address"] = context["theme_address"].strip().split("\n")

    if isinstance(context["theme_social"], str):
        pairs = [p.split(",") for p in context["theme_social"].strip().split("\n")]
        social = [{k: v.strip() for k, v in zip(("type", "link"), p)} for p in pairs]
        # we need to update the constructed context
        context["config"]["extra"]["social"] = social

    if "copyright" in context and context["copyright"]:
        logger.warning(
            "you have defined 'copyright' in your conf.py, "
            "the default Blue Brain project copyright will be used."
        )

    context["config"][
        "copyright"
    ] = f"&copy; Blue Brain Project/EPFL 2005-{datetime.datetime.now().year}. All rights reserved."



"""This module contains the classes used to build the search_index.json file for lunr search.

See https://github.com/mkdocs/mkdocs/blob/master/mkdocs/contrib/search/search_index.py
which shows the expected structure of the JSON search index and the
approach used by mkdocs to generate it
"""

import re
from json import dumps

try:
    from html.parser import HTMLParser
except ImportError:
    from HTMLParser import HTMLParser


class IndexEntry:
    """Holds the data for each entry in the search index."""

    MULTI_SPACE_REGEX = re.compile(r"\s+")

    def __init__(self, location="", title=""):
        """Initialize the index entry."""
        self.location = location
        self.title = title
        self.text_list = []

    @property
    def text(self):
        """Generates the search index text for on an entry.

        Generates the joined and cleaned text for the entry from the list
        of text found in the page/section of the entry.
        """
        text = "\n".join(self.text_list)
        text = text.replace("\u00B6", " ")  # remove the pilcrow symbol
        text = self.MULTI_SPACE_REGEX.sub(" ", text).strip()

        return text

    def as_dict(self):
        """Returns a dict containing the data required for the search index, used by json.dumps."""
        return {k: getattr(self, k) for k in ("location", "title", "text")}


# pylint: disable=abstract-method
class SearchIndexBuilder(HTMLParser):
    """Class is used to generate the search_index.json file.

    It is intended to hold all the data required and then be dumped
    into a jinja templated file using __repr__.
    """

    HEADING_TAGS = {"h" + str(i) for i in range(1, 7)}

    def __init__(self):
        """Initialize the parser and additional attributes."""
        HTMLParser.__init__(self)
        self._entries = []
        self._current_page = None
        self._current_section = None
        self._current_section_id = None
        self._current_section_id = None
        self._is_heading = False

    def add_page_to_index(self, page):
        """Adds a single page to the search index.

        This takes care of creating the entry for the complete page as well as splitting it into
        each section.
        """
        # the path needs to move to the parent directory as the index is
        # stored in the _static directory
        self._current_page = IndexEntry("../" + page.url, page.title)
        self._entries.append(self._current_page)

        contents = page.content or ""
        self.feed(contents)

    def __repr__(self):
        """Returns the dumped json as the repr for the entry.

        The dumped json is used for the search index build.
        """
        return dumps(
            {"docs": [e.as_dict() for e in self._entries], "config": {}}, sort_keys=True
        )

    def handle_starttag(self, tag, attrs):
        """Handles the possible HTML tags as required for building the index."""
        attrsdict = dict(attrs)

        # sphinx puts each section inside a div which actually has the
        # id for the section. this is changed with JS for the final
        # page rendering, but at this point we need to get the id
        # from the enclosing div.
        if tag == "div" and "section" in attrsdict.get("class", ""):
            self._current_section_id = attrsdict.get("id", None)
            return

        if tag in self.HEADING_TAGS and self._current_section_id is not None:
            # add the anchor for the section to the page url
            section_location = "#".join(
                (self._current_page.location, self._current_section_id)
            )
            self._current_section = IndexEntry(section_location)
            self._entries.append(self._current_section)
            self._is_heading = True
            self._current_section_id = None
        else:
            self._is_heading = False

    def handle_endtag(self, tag):
        """Handles closing tags, particularly headings."""
        if tag in self.HEADING_TAGS and self._current_section_id is not None:
            self._is_heading = False
            self._current_section_id = None

    def handle_data(self, data):
        """Handles appending data to the correct place (heading or section)."""
        # always add the data to the complete page entry
        self._current_page.text_list.append(data)

        if self._is_heading:
            self._current_section.title = data
        elif self._current_section is not None:
            self._current_section.text_list.append(data)



"""Override autodoc output to wrap in a div."""

from docutils import nodes
import sphinx.ext.autodoc as sea
from sphinx.ext.autodoc.directive import AutodocDirective


class AutodocDirectiveOverride(AutodocDirective):
    """Extends AutodocDirective to wrap output in a div."""

    def run(self):
        """Wrap the autodoc output in a div with autodoc class."""
        result = super().run()
        container = nodes.container()
        container["classes"].append("autodoc-output")
        container += result
        return [container]


def add_autodoc_override(app):
    """Override the autodoc definitions to wrap them in a div."""
    if "sphinx.ext.autodoc" in app.extensions:
        # this will override the normal autodoc directive if required. See:
        # https://github.com/sphinx-doc/sphinx/blob/master/sphinx/ext/autodoc/__init__.py#L1464
        # https://github.com/sphinx-doc/sphinx/blob/15bc5a32bb0b78b432803e55fffa10c801182c75/sphinx/application.py#L996
        documenters = [
            sea.ModuleDocumenter,
            sea.ClassDocumenter,
            sea.ExceptionDocumenter,
            sea.DataDocumenter,
            sea.FunctionDocumenter,
            sea.DecoratorDocumenter,
            sea.MethodDocumenter,
            sea.AttributeDocumenter,
            sea.InstanceAttributeDocumenter,
        ]

        for d in documenters:
            app.add_directive(
                "auto" + d.objtype, AutodocDirectiveOverride, override=True
            )



"""Module of utilities for translation between mkdocs and Sphinx."""

from sphinx_bluebrain_theme.utils.toc_builder import build_tocs
from sphinx_bluebrain_theme.utils.inject_context import inject_context_variables
from sphinx_bluebrain_theme.utils.filters import add_filters
from sphinx_bluebrain_theme.utils.autodoc_override import add_autodoc_override
from sphinx_bluebrain_theme.utils.metadata import write_metadata, write_metadata_sphinx



"""This extension provides collapsible areas on an html page."""

from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.util.docutils import SphinxDirective


class details_node(nodes.General, nodes.Element):
    """Details node."""


def visit_details_node(self, node):
    """Create the opening details node html."""
    attrs = {}

    if node.get("open", False):
        attrs["open"] = ""

    self.body.append(
        self.starttag(node, "details", CLASS=" ".join(node["classes"]), **attrs)
    )


def depart_details_node(self, node):
    """Close the details html."""
    del node
    self.body.append("</details>")


class summary_node(nodes.General, nodes.Element):
    """Summary node."""


def visit_summary_node(self, node):
    """Create the opening summary node html."""
    self.body.append(self.starttag(node, "summary", "", CLASS=""))


def depart_summary_node(self, node):
    """Close the summary html."""
    del node
    self.body.append("</summary>")


class Details(SphinxDirective):
    """A container for details/summary."""

    has_content = True
    required_arguments = 1
    final_argument_whitespace = True
    option_spec = {"class": directives.class_option, "open": directives.flag}

    def run(self):
        """Insert correct nodes for directive."""
        self.assert_has_content()

        node = details_node()

        node["classes"] += self.options.get("class", [])
        node["open"] = "open" in self.options

        summary = summary_node()
        summary += nodes.Text(self.arguments[-1])
        node += summary

        self.add_name(node)

        self.state.nested_parse(self.content, self.content_offset, node)

        return [node]


def setup(app):
    """Register nodes and directives for use."""
    app.add_node(details_node, html=(visit_details_node, depart_details_node))
    app.add_node(summary_node, html=(visit_summary_node, depart_summary_node))
    app.add_directive("details", Details)



"""sphinx_bluebrain_theme package."""

from os import path

from sphinx_bluebrain_theme import utils


def setup(app):
    """Initialise the theme and connect theme specific functions to events."""
    app.add_html_theme("sphinx-bluebrain-theme", path.dirname(path.abspath(__file__)))
    app.setup_extension("sphinx_bluebrain_theme.ext.tabs")
    app.setup_extension("sphinx_bluebrain_theme.ext.details")
    app.setup_extension("sphinx_bluebrain_theme.ext.iframe")
    app.connect("builder-inited", utils.add_filters)
    app.connect("builder-inited", utils.add_autodoc_override)
    app.connect("env-updated", utils.write_metadata_sphinx)
    app.connect("html-page-context", utils.build_tocs)
    app.connect("html-page-context", utils.inject_context_variables)



"""Classes and functions used to convert Sphinx generated Tables of Contents.

The original theme was built for mkdocs and so expects mkdocs classes to be
provided to templates for rendering. These classes do the job of converting
the HTML table of contents generated by Sphinx into a format compatible
with the theme.

mkdocs ToC HTML parser is built for their specific markdown output. See here:
- https://github.com/mkdocs/mkdocs/blob/master/mkdocs/structure/toc.py
"""

try:
    from html.parser import HTMLParser
except ImportError:
    from HTMLParser import HTMLParser

from sphinx_bluebrain_theme.utils.navutils import Nav, Page


# pylint: disable=abstract-method
class TocHTMLParser(HTMLParser):
    """Customised Sphinx Table of Contents HTML parser.

    This parser will construct the table of contents from the HTML produced by Sphinx.
    """

    def __init__(self, allow_parent_links=False):
        """Initialize the ToC parser."""
        HTMLParser.__init__(self)

        self.nav = Nav()
        self._current_item = None
        self._use_data = False
        self._parent_stack = [self.nav]
        self._is_first_ul = True
        self._allow_parent_links = allow_parent_links

    def reset(self):
        """Reset the HTML parser."""
        HTMLParser.reset(self)

        self.nav = Nav()
        self._current_item = None
        self._use_data = False
        self._parent_stack = [self.nav]
        self._is_first_ul = True

    def handle_starttag(self, tag, attrs):
        """Called when the parser encounters a new opening tag.

        Args:
            tag (str): The name of the tag.
            attrs (tuple): The HTML attributes of the tag.
        """
        if tag == "ul":
            if self._is_first_ul:
                self._is_first_ul = False
            else:
                # a new list, so we push the current item on the parent stack to
                # prepare to add children to it
                self._parent_stack.append(self._current_item)
                if not self._allow_parent_links:
                    self._current_item.url = None
        elif tag == "li":
            # this is a child item, so create a new item and append it to the
            # current parent

            self._current_item = Page()
            self._current_item.parent = self._parent_stack[-1]
            self._parent_stack[-1].children.append(self._current_item)

            # find the active page
            for attr, data in attrs:
                if attr == "class":
                    self._current_item.active = "current" in data
        elif tag == "a":
            # get the url associated and make sure we will capture the next
            # data element, it is the text of the ToC entry
            for attr, data in attrs:
                if attr == "href":
                    self._current_item.url = data
            self._use_data = True

    def handle_endtag(self, tag):
        """Called when the parser encounters a closing tag.

        Args:
            tag (str): The name of the tag which is closed.
        """
        # we've reached the end of a list, and the end of the current item
        if tag == "ul":
            self._current_item = self._parent_stack.pop()

    def handle_data(self, data):
        """Called when the parser encounters data enclosed by a tag.

        Args:
            data (str): The data enclosed by a tag.
        """
        # if we need the link text, get it
        if self._use_data:
            self._current_item.title = data
            self._use_data = False


# pylint: disable=unused-argument
def build_tocs(app, pagename, templatename, context, doctree):
    """Injects the global and local Tables of Contents into the HTML context.

    Args:
        app: The Sphinx application.
        pagename (str): The name of the current page.
        templatename (str): The name of the template used to render the page.
        context (dict): The context passed to the page template for rendering.
        doctree: A doctree when the page is created from reST documents.
    """
    # build the global table of contents
    if "toctree" in context:
        global_toc_html = context["toctree"](
            maxdepth=-1, collapse=False, includehidden=True, titles_only=True
        )
        parser = TocHTMLParser()
        parser.feed(global_toc_html)
        homepage_url = context["pathto"](context["master_doc"], 1)
        parser.nav.set_homepage(homepage_url)
        context["globaltoc"] = parser.nav

    # build the local table of contents
    if "toc" in context:
        parser = TocHTMLParser(allow_parent_links=True)
        parser.feed(context["toc"])

        # check there is a heading and only set the local ToC if there is
        if len(parser.nav):
            context["localtoc"] = parser.nav[0]



"""Utilities for creating a metadata file."""
import datetime
import json
import os
import sys

from contextlib import contextmanager
from subprocess import check_output, CalledProcessError

import sphinx
from pkg_resources import get_distribution


logger = sphinx.util.logging.getLogger(__name__)

METADATA_NAMES = ("METADATA", "PKG-INFO")


@contextmanager
def change_cwd(temp_cwd):
    """Temporarily change the current working directory."""
    original_cwd = os.getcwd()
    try:
        os.chdir(temp_cwd)
        yield
    finally:
        os.chdir(original_cwd)


@contextmanager
def add_to_path(temp_path):
    """Temporarily add a path to the system PATH."""
    # use list() to copy
    path_original = list(sys.path)
    try:
        sys.path.insert(0, temp_path)
        yield
    finally:
        sys.path = path_original


def write_metadata(metadata, output_dir):
    """Write a metadata dict to a directory."""
    metadata_path = os.path.join(output_dir, "metadata.md")
    with open(metadata_path, "w", encoding="utf8") as fd:
        fd.write("---\n")
        for k, v in metadata.items():
            fd.write(str(k) + ": " + str(v) + "\n")
        fd.write("---\n")

    metadata_json_path = os.path.join(output_dir, "metadata.json")
    with open(metadata_json_path, "w", encoding="utf8") as fd:
        fd.write(json.dumps(metadata))

    return metadata_path


def get_metadata_from_json(json_path):
    """Get metadata from a json file.

    The file must follow the standard structure of package.json files.
    """
    with open(json_path, encoding="utf8") as fd:
        metadata = json.load(fd)

    output = {
        "name": metadata["name"],
        "version": metadata["version"],
        "description": metadata["description"],
        "homepage": metadata["homepage"],
        "repository": metadata["repository"]["url"],
        "issuesurl": metadata["bugs"]["url"],
        "license": metadata["license"],
        "maintainers": metadata["author"],
    }

    # contributors is a special case where they might be
    # specified in the conf.py so don't raise an error
    output["contributors"] = ",".join(metadata.get("contributors", ""))

    return output


def build_metadata_from_setuptools_dict(metadata):
    """Build and return metadata from a setuptools dict.

    This is typically from package metadata.
    """
    # based on warehouse and virtualenv examples
    # https://github.com/pypa/warehouse/blob/master/warehouse/templates/packaging/detail.html
    # https://github.com/pypa/virtualenv/blob/master/setup.cfg
    project_urls = metadata["project_urls"]
    output = {
        "name": metadata["name"],
        "version": metadata["version"],
        "description": metadata["description"],
        "homepage": metadata["url"],
        "license": metadata["license"],
        "maintainers": metadata["author"],
        "repository": project_urls["Source"],
        "issuesurl": project_urls["Tracker"],
    }

    # contributors is a special case where they might be
    # specified in the conf.py so don't raise an error
    output["contributors"] = metadata.get("contributors", None)

    return output


def get_metadata_from_distribution(distribution_name):
    """Get the metadata from a distribution."""
    # useful information: https://packaging.python.org/specifications/core-metadata/
    dist = get_distribution(distribution_name)

    metadata = {}

    # get metadata name or error
    for mdn in METADATA_NAMES:
        if dist.has_metadata(mdn):
            metadata_name = mdn
            break
    else:
        raise FileNotFoundError(
            f"package has no metadata file with name: {METADATA_NAMES}"
        )

    for mdl in dist.get_metadata_lines(metadata_name):
        # guard against silly data
        if ":" not in mdl:
            continue

        key, value = mdl.split(":", 1)
        key = key.lower()
        value = value.strip()

        # treat UNKNOWN as no value, this the setuptools metadata
        # equivalent of None
        if value == "UNKNOWN":
            continue

        # handle special cases
        if key == "project-url":
            key = "project_urls"
            inner_key, inner_value = value.split(",", 1)
            inner_value = inner_value.strip()
            # project-urls is a dict
            value = metadata.get(key, {})
            value[inner_key] = inner_value
        elif key in {
            "platform",
            "supported-platform",
            "requires-dist",
            "requires-external",
            "provides-extra",
            "provides-dist",
            "obsoletes-dist",
            "classifier",
        }:
            key = key.replace("-", "_") + "s"
            inner_value = value
            # these keys have list values
            value = metadata.get(key, [])
            value.append(inner_value)

        metadata[key] = value

    # allow summary as the description
    if "description" not in metadata:
        metadata["description"] = metadata["summary"]

    # home-page needs to be used for the url
    if "home-page" in metadata:
        metadata["url"] = metadata["home-page"]

    return build_metadata_from_setuptools_dict(metadata)


def build_metadata(metadata_file=None, metadata_overrides=None, distribution_name=None):
    """Read metadata from sources."""
    metadata = {}

    # load metadata from file if specified
    if distribution_name is not None:
        metadata_package = get_metadata_from_distribution(distribution_name)
        metadata.update(metadata_package)
    elif metadata_file:
        if not os.path.isfile(metadata_file):
            raise FileNotFoundError(
                f"{metadata_file}: specified metadata_file does not exist"
            )

        _, ext = os.path.splitext(metadata_file)
        if ext != ".json":
            raise ValueError(f"{metadata_file}: expected a .json file")

        metadata_package = get_metadata_from_json(metadata_file)
        metadata.update(metadata_package)

    if metadata_overrides:
        metadata.update(metadata_overrides)

    # early return if no metadata
    if not metadata:
        return None

    # add updated key if not there already
    if "updated" not in metadata:
        metadata["updated"] = datetime.datetime.now().strftime("%d/%m/%y")

    return metadata


def write_metadata_sphinx(app, exception):  # pylint: disable=unused-argument
    """Write the metadata to the output directory."""
    # get the metadata file from context variable
    metadata_file = app.config["html_theme_options"].get("metadata_file", None)
    metadata_dist = app.config["html_theme_options"].get("metadata_distribution", None)

    if metadata_dist is not None:
        if metadata_file:
            raise ValueError("only one of metadata_file and metadata_dist is allowed")

        metadata_file = None

    if metadata_file is not None and not os.path.isabs(metadata_file):
        metadata_file = os.path.abspath(os.path.join(app.confdir, metadata_file))

    # load extra metadata from the given context variable
    metadata_overrides = app.config["html_theme_options"].get("metadata_overrides", {})

    # get the path to the git repo if given
    repo_path = app.config["html_theme_options"].get("repo_path", None)

    metadata = build_metadata(metadata_file, metadata_overrides, metadata_dist)

    if not metadata:
        return

    # confirm that versions match between metadata and sphinx
    # if using a distribution for metadata, it will catch the case
    # that another version of the same package has been imported
    assert (
        metadata["version"] == app.config["version"]
    ), "conf.py version and metadata version do not match"

    contributors = metadata.get("contributors", None)

    # if contributors is a git repo, get the top five
    if contributors is None:
        clone_path = app.confdir
        if repo_path is not None:
            if not os.path.isabs(repo_path):
                repo_path = os.path.abspath(os.path.join(app.confdir, repo_path))
            if not (os.path.isdir(repo_path) or os.path.isfile(repo_path)):
                raise IOError(f"{repo_path}: not a git repository")
            clone_path = os.path.dirname(repo_path)

        # check if it is a git repo
        try:
            cmd = ["git", "-C", clone_path, "rev-parse", "--is-inside-work-tree"]
            is_in_git_repo = check_output(cmd).decode("utf-8") == "true"
        except CalledProcessError:
            is_in_git_repo = False

        if is_in_git_repo:
            cmd = ["git", "-C", clone_path, "shortlog", "-sne", "origin/master"]

            contributors = check_output(cmd).decode("utf-8").splitlines()
            contributors = [line.split("\t", 2)[1] for line in contributors]
            has_etal = len(contributors) > 5

            # get top five contributors and account for any additional
            contributors = ", ".join(contributors[:5]) + " et al." if has_etal else ""
        elif not repo_path:
            contributors = "None"

        metadata["contributors"] = contributors

    # don't check for None here, as we also want to catch an empty string
    if not metadata.get("contributors", None):
        raise ValueError("contributors must be provided or repo_path given")

    # write the metadata
    write_metadata(metadata=metadata, output_dir=app.outdir)



"""This extension provides tabbed code areason an html page."""

from docutils import nodes
from sphinx.util.docutils import SphinxDirective
from sphinx.directives.code import CodeBlock, LiteralInclude


class CodeTabMixin:
    """Common code for code-block and literal-include type tabs."""

    def run_mixin(self, inner):
        """Insert correct nodes for code tabs."""
        # tab title is the second arg
        title = self.arguments.pop()

        # tab_ids start from 1, so increment at start
        self.env.tab_group_tab_id += 1

        # build the tab_id
        group_name = self.env.tab_group_name
        tab_id_int = self.env.tab_group_tab_id
        tab_id = f"{group_name}_{tab_id_int}"

        node = nodes.container()
        node["classes"].append("superfences-content")

        self.add_name(node)

        # include the node for the actual code
        node += inner

        # create the input node for the tab
        inp = tab_node(ids=[tab_id])
        inp.tagname = "input"
        inp["attrs"] = {"type": "radio", "name": group_name}
        if self.env.tab_group_tab_id == 1:
            inp["attrs"]["checked"] = "checked"

        # create the label
        lab = tab_node()
        lab.tagname = "label"
        lab["attrs"] = {"for": tab_id}

        # add text to the label
        txt = nodes.Text(title)
        lab += txt

        return [inp, lab, node]


class CodeTab(CodeTabMixin, CodeBlock):
    """Code tab convenience class to avoid extra directives."""

    required_arguments = 2

    def run(self):
        """Run the CodeTab including mixin."""
        inner = super().run()
        return self.run_mixin(inner)


class LiteralTab(CodeTabMixin, LiteralInclude):
    """Literal include convenience class."""

    required_arguments = 2

    def run(self):
        """Run the LiteralTab including mixin."""
        inner = super().run()
        return self.run_mixin(inner)


class tab_node(nodes.General, nodes.Element):
    """Custom tab node that can be used for different tags."""


def visit_tab_node(self, node):
    """Create the opening tab node html including data attributes."""
    attrs = node.get("attrs", {})
    self.body.append(self.starttag(node, node.tagname, "", CLASS="", **attrs))


def depart_tab_node(self, node):
    """Close the tab html."""
    self.body.append(f"</{node.tagname}>")


class TabGroup(SphinxDirective):
    """A container for multiple tabs."""

    has_content = True
    required_arguments = 0
    optional_arguments = 0

    def run(self):
        """Insert correct nodes for directive."""
        self.assert_has_content()
        text = "\n".join(self.content)

        self.env.tab_group_id = getattr(self.env, "tab_group_id", 0) + 1
        self.env.tab_group_name = f"__tab_{self.env.tab_group_id}"
        self.env.tab_group_tab_id = 0
        node = nodes.container(text)
        node["classes"].append("superfences-tabs")

        self.state.nested_parse(self.content, self.content_offset, node)

        return [node]


def setup(app):
    """Register nodes and directives for use."""
    app.add_node(tab_node, html=(visit_tab_node, depart_tab_node))
    app.add_directive("codetab", CodeTab)
    app.add_directive("literaltab", LiteralTab)
    app.add_directive("tabgroup", TabGroup)



"""This extension provides iframes for embedding html."""

from docutils import nodes
from docutils.parsers.rst import directives
from sphinx.util.docutils import SphinxDirective


class iframe_node(nodes.General, nodes.Element):
    """iframe node."""


def visit_iframe_node(self, node):
    """Create the opening iframe node html."""
    attrs = node.get("attrs", {})

    self.body.append(
        self.starttag(node, "iframe", CLASS=" ".join(node["classes"]), **attrs)
    )


def depart_iframe_node(self, node):
    """Close the iframe html."""
    del node
    self.body.append("</iframe>")


class Iframe(SphinxDirective):
    """A container for iframe."""

    has_content = False
    required_arguments = 1
    final_argument_whitespace = False
    option_spec = {"class": directives.class_option}

    def run(self):
        """Insert correct nodes for directive."""
        node = iframe_node()

        node["classes"] += ["sphinx-iframe"]
        node["classes"] += self.options.get("class", [])

        node["attrs"] = {"src": self.arguments[0], "scrolling": "no"}

        self.add_name(node)

        self.state.nested_parse(self.content, self.content_offset, node)

        return [node]


def setup(app):
    """Register nodes and directives for use."""
    app.add_node(iframe_node, html=(visit_iframe_node, depart_iframe_node))
    app.add_directive("iframe", Iframe)



"""Utilities for building mkdocs compatible navigation objects.

Since the original theme is built for mkdocs, all the templates
expect mkdocs classes (and their attributes), so these classes
provide compatible data structures.

The classes are mainly data holders with a very few
convenience methods to assist with building and maintaining
consistent navigation structures.

See mkdocs description, which lists all the expected attributes:
- https://www.mkdocs.org/user-guide/custom-themes/#nav
- https://www.mkdocs.org/user-guide/custom-themes/#pages

See also mkdocs implementation:
- https://github.com/mkdocs/mkdocs/tree/master/mkdocs/structure
"""


class NavBase(object):  # pylint: disable=useless-object-inheritance
    """Base class for navigation objects."""

    is_page = False

    def __init__(self):
        """Initialize the NavBase object."""
        self.title = None
        self.parent = None
        self.children = []
        self._active = False

    def __iter__(self):
        """Iterates over the object's children."""
        return iter(self.children)

    def __len__(self):
        """Returns the number of children."""
        return len(self.children)

    def __getitem__(self, key):
        """Returns the nth child."""
        return self.children[key]

    @property
    def active(self):
        """Return whether this nav object is active."""
        return self._active

    @active.setter
    def active(self, value):
        """Set the active state of the nav object also setting its parent to active."""
        is_active = bool(value)
        self._active = is_active

        # do not allow making the parent inactive
        if self.parent is not None and is_active:
            self.parent.active = is_active


class Nav(NavBase):
    """Class for top level nav object, container for all others."""

    def __init__(self):
        """Initialize the base class and Nav specific attributes."""
        super().__init__()
        self.homepage = None
        self.pages = []
        self.title = "Base Nav"

    def set_homepage(self, homepage_name):
        """Search for, and set, the homepage."""
        # homepage should be in the top level of pages
        for page in self.children:
            if page.url in (homepage_name + ".html", "#"):
                page.is_homepage = True
                return


class Page(NavBase):
    """Dedicated class for Page."""

    is_page = True

    def __init__(self):
        """Initialize the base class and Page specific attributes."""
        super().__init__()
        self.content = None
        self.toc = None
        self.meta = None
        self.url = None
        self.abs_url = None
        self.canonical_url = None
        self.edit_url = None
        self.is_homepage = False
        self.previous_page = None
        self.next_page = None
        self.ancestors = None

    def __eq__(self, other):
        """Evaluates equality based on active status only."""
        return isinstance(other, Page) and self.active and other.active

    def __bool__(self):
        """If the page has a title and url then it is true."""
        return self.title is not None and self.url is not None
